#!/usr/bin/env python

from pcaspy import Driver, SimpleServer

prefix = 'MTEST:'
pvdb = {
    'LEVEL': {
    },
    'FILL' : {
        'prec' : 3,
        'asg'  : 'fill',
    },
}

class myDriver(Driver):
    def  __init__(self):
        super(myDriver, self).__init__()

if __name__ == '__main__':
    server = SimpleServer()
    server.initAccessSecurityFile('test.as', P=prefix)
    server.createPV(prefix, pvdb)
    driver = myDriver()
    # process CA transactions
    while True:
        server.process(0.1)



#
# macro definition from alarm.h

class Severity:
    NO_ALARM        = 0
    MINOR_ALARM     = 1
    MAJOR_ALARM     = 2
    INVALID_ALARM   = 3

    @staticmethod
    def nameOf(index):
        return SeverityStrings[index]

class Alarm:
    NO_ALARM        = 0
    READ_ALARM      = 1
    WRITE_ALARM     = 2
    HIHI_ALARM      = 3
    HIGH_ALARM      = 4
    LOLO_ALARM      = 5
    LOW_ALARM       = 6
    STATE_ALARM     = 7
    COS_ALARM       = 8
    COMM_ALARM      = 9
    TIMEOUT_ALARM   =10
    HW_LIMIT_ALARM  =11
    CALC_ALARM      =12
    SCAN_ALARM      =13
    LINK_ALARM      =14
    SOFT_ALARM      =15
    BAD_SUB_ALARM   =16
    UDF_ALARM       =17
    DISABLE_ALARM   =18
    SIMM_ALARM      =19
    READ_ACCESS_ALARM =20
    WRITE_ACCESS_ALARM=21

    @staticmethod
    def nameOf(index):
        return AlarmStrings[index]
#
# string representation from alarmString.h

SeverityStrings = [
    "NO_ALARM",
    "MINOR",
    "MAJOR",
    "INVALID"
]

AlarmStrings = [
    "NO_ALARM",
    "READ",
    "WRITE",
    "HIHI",
    "HIGH",
    "LOLO",
    "LOW",
    "STATE",
    "COS",
    "COMM",
    "TIMEOUT",
    "HWLIMIT",
    "CALC",
    "SCAN",
    "LINK",
    "SOFT",
    "BAD_SUB",
    "UDF",
    "DISABLE",
    "SIMM",
    "READ_ACCESS",
    "WRITE_ACCESS"
]


#!/usr/bin/env python

from pcaspy import Driver, SimpleServer

prefix = 'MTEST:'
pvdb = {
    'RAND' : {
        'prec' : 3,
    },
}

class myDriver(Driver):
    def  __init__(self):
        super(myDriver, self).__init__()

if __name__ == '__main__':
    server = SimpleServer()
    server.createPV(prefix, pvdb)
    driver = myDriver()

    # process CA transactions
    while True:
        server.process(0.1)



#!/usr/bin/env python
import threading
import numpy

from pcaspy import Driver, SimpleServer, Severity

MAX_POINTS      = 1000
FREQUENCY       = 1000
AMPLITUDE       = 1.0
NUM_DIVISIONS   = 10
MIN_UPDATE_TIME = 0.01

prefix = 'MTEST:'
pvdb = {
        'Run'              : { 'type' : 'enum',
                               'enums': ['STOP', 'RUN'],
                               'states': [Severity.NO_ALARM, Severity.MINOR_ALARM]},
        'MaxPoints'        : { 'type' : 'int' ,
                               'value': MAX_POINTS},
        'UpdateTime'       : { 'prec' : 3, 'unit' : 's', 'value' : 1     },
        'TimePerDivision'  : { 'prec' : 5, 'value' : 0.001 },
        'TriggerDelay'     : { 'prec' : 5, 'unit' : 's', 'value' : 0.0005},
        'VoltsPerDivision' : { 'prec' : 3, 'value' : 0.2   },
        'VoltOffset'       : { 'prec' : 3, 'unit' : 'V' },
        'NoiseAmplitude'   : { 'prec' : 3, 'value' : 0.2   },
        'Waveform'         : { 'count': MAX_POINTS,
                                'prec' : 5 },
        'TimeBase'         : { 'count': MAX_POINTS,
                               'prec' : 5,
                               'value': numpy.arange(MAX_POINTS, dtype=float) 
                                            * NUM_DIVISIONS / (MAX_POINTS -1)},
        'MinValue'         : { 'prec' : 4 },
        'MaxValue'         : { 'prec' : 4 },
        'MeanValue'        : { 'prec' : 4 },
}
 
class myDriver(Driver):
    def __init__(self):
        Driver.__init__(self)
        self.eid = threading.Event()
        self.tid = threading.Thread(target = self.runSimScope) 
        self.tid.setDaemon(True)
        self.tid.start()

    def write(self, reason, value):
        status = True
        # take proper actions
        if reason == 'UpdateTime':
            value = max(MIN_UPDATE_TIME, value)
        elif reason == 'Run':
            if not self.getParam('Run') and value == 1:
                self.eid.set()
                self.eid.clear()
        # store the values
        if status:
            self.setParam(reason, value)
        return status

    def runSimScope(self):
        # simulate scope waveform
        while True:
            run = self.getParam('Run')
            updateTime = self.getParam('UpdateTime')
            if run:
                self.eid.wait(updateTime)
            else:
                self.eid.wait()
            run = self.getParam('Run')
            if not run: continue
            # retrieve parameters
            noiseAmplitude    = self.getParam('NoiseAmplitude')
            timePerDivision   = self.getParam('TimePerDivision')
            voltsPerDivision  = self.getParam('VoltsPerDivision')
            triggerDelay      = self.getParam('TriggerDelay')
            voltOffset        = self.getParam('VoltOffset')
            # calculate the data wave based on timeWave scale 
            timeStart = triggerDelay
            timeStep  = timePerDivision * NUM_DIVISIONS / MAX_POINTS
            timeWave  = timeStart + numpy.arange(MAX_POINTS) * timeStep 
            noise  = noiseAmplitude * numpy.random.random(MAX_POINTS)
            data = AMPLITUDE * numpy.sin(timeWave * FREQUENCY * 2 * numpy.pi) + noise 
            # calculate statistics
            self.setParam('MinValue',  data.min())
            self.setParam('MaxValue',  data.max())
            self.setParam('MeanValue', data.mean())
            # scale/offset
            yScale = 1.0 / voltsPerDivision
            data   = NUM_DIVISIONS/2.0 + yScale * (data + voltOffset)
            self.setParam('Waveform',  data)
            # do updates so clients see the changes
            self.updatePVs()

if __name__ == '__main__':
    server = SimpleServer()
    server.createPV(prefix, pvdb)
    driver = myDriver()

    # process CA transactions
    while True:
        server.process(0.1)



import sys
from PyQt4 import QtGui, QtCore

from pcaspy import SimpleServer, Driver
from pcaspy.tools import ServerThread

prefix = 'MTEST:'
pvdb   = {
        'RAND': {}
}

class myDriver(Driver):
    def __init__(self):
        super(myDriver,self).__init__()

class Display(QtGui.QWidget):
    def __init__(self):
        super(Display, self).__init__()
        layout = QtGui.QHBoxLayout()
        layout.addWidget(QtGui.QLabel('Value:'))
        input = QtGui.QDoubleSpinBox()
        layout.addWidget(input)
        self.setLayout(layout)
        self.connect(input, QtCore.SIGNAL('valueChanged(double)'), self.newValue)
        self.drv = myDriver()

    def newValue(self, value):
        self.drv.write('RAND', value)
        self.drv.updatePVs()

if __name__ == '__main__':
    # create pcas server
    server = SimpleServer()
    server.createPV(prefix, pvdb)
    
    # create qt gui
    app = QtGui.QApplication(sys.argv)
    win = Display()
    win.show()
    
    # create pcas server thread and shut down when app exits
    server_thread = ServerThread(server)
    QtCore.QObject.connect(app, QtCore.SIGNAL('lastWindowClosed()'), server_thread.stop)

    # start pcas and gui event loop
    server_thread.start()
    app.exec_()



from . import cas
from .alarm import Severity, Alarm
import collections
import operator
import threading
import time
import sys
import logging
if sys.hexversion >= 0x02070000:
    from logging import NullHandler
else:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger('pcaspy').addHandler(NullHandler())


class Manager(object):
    pvs = {}     #: PV dict using port name as key and {pv base name: pv instance} as value
    pvf = {}     #: PV dict using PV full name as key
    driver = {}  #: Driver dict

# Yes, this is a global instance
manager = Manager()


# decorator to register driver
def registerDriver(driver_init_func):
    def wrap(*args, **kargs):
        driver_instance = args[0]
        port = driver_instance.port
        driver_init_func(*args, **kargs)
        manager.driver[port] = driver_instance
    return wrap


# Driver metaclass to decorate subclass.__init__ to
# register subclass object
class DriverType(type):
    def __init__(cls, name, bases, dct):
        if name != 'Driver':
            cls.__init__ = registerDriver(cls.__init__)
        type.__init__(cls, name, bases, dct)


class Data(object):
    def __init__(self):
        self.value = 0
        self.flag = False
        self.severity = Severity.INVALID_ALARM
        self.alarm = Alarm.UDF_ALARM
        self.udf = True
        self.mask = 0
        self.time = cas.epicsTimeStamp()

    def __repr__(self):
        return "value=%s alarm=%s severity=%s flag=%s mask=%s time=%s" % \
               (self.value, Alarm.nameOf(self.alarm), Severity.nameOf(self.severity), self.flag, self.mask, self.time)

# Define empty DriverBase using metaclass syntax compatible with both Python 2 and Python 3
DriverBase = DriverType(str('DriverBase'), (), {
    '__doc__': 'Driver base class'
})


class Driver(DriverBase):
    """
    This class reacts to PV's read/write requests. The default behavior is to accept any value of a write request
    and return it to a read request, an echo alike.

    To specify the behavior, override methods :meth:`read` and :meth:`write` in a derived class.
    """
    port = 'default'

    def __init__(self):
        """
        Initialize parameters database. This method must be called by subclasses in the first place.
        """
        self.pvDB = {}
        # init pvData with pv instance
        for reason, pv in manager.pvs[self.port].items():
            data = Data()
            data.value = pv.info.value
            self.pvDB[reason] = data

    def read(self, reason):
        """
        Read PV current value

        :param str reason: PV base name
        :return: PV current value

        This method is invoked by server library when clients issue read access to a PV.
        By default it returns the value stored in the parameter library by calling :meth:`getParam`.

        The derived class might leave this method untouched and update the PV values from
        a separate polling thread. See :ref:`shell-command-example`, :ref:`simscope-example`.

        .. note:: This method is called by the server library main thread. Time consuming tasks
                  should not be performed here. It is suggested to work in an auxiliary thread.
        """
        return self.getParam(reason)

    def write(self, reason, value):
        """
        Write PV new value

        :param str reason: PV base name
        :param value: PV new value
        :return: True if the new value is accepted, False if rejected.

        This method is invoked by server library when clients write to a PV.
        By default it stores the value in the parameter library by calling :meth:`setParam`.

        .. note:: This method is called by the server library main thread. Time consuming tasks
                  should not be performed here. It is suggested to work in an auxiliary thread.

        """
        self.setParam(reason, value)
        return True

    def setParam(self, reason, value):
        """set PV value and request update

        :param str reason: PV base name
        :param value: PV new value

        Store the PV's new value if it is indeed different from the old.
        For list and numpy array, a copy will be made.
        This new value will be pushed to registered client the next time when :meth:`updatePVs` is called.
        The timestamp will be updated to the current time anyway.

        Alarm and severity status are updated as well. For numeric type, the alarm/severity is determined as the
        following:

            ========================    ============  ============
            value                       alarm         severity
            ========================    ============  ============
            value < *lolo*              LOLO_ALARM    MAJOR_ALARM
            *lolo* < value < *low*      LOW_ALARM     MINOR_ALARM
            *low* < value < *high*      NO_ALARM      NO_ALARM
            *high* < value < *hihi*     HIGH_ALARM    MINOR_ALARM
            value > *hihi*              HIHI_ALARM    MAJOR_ALARM
            ========================    ============  ============

        For enumerate type, the alarm severity is defined by field *states*. And if severity is other than NO_ALARM,
        the alarm status is STATE_ALARM.

        """
        # make a copy of mutable objects, list, numpy.ndarray
        if isinstance(value, list):
            value = value[:]
        elif 'numpy.ndarray' in str(type(value)):
            value = value.copy()
        # check whether value update is needed
        pv = manager.pvs[self.port][reason]
        self.pvDB[reason].mask |= pv.info.checkValue(value)
        self.pvDB[reason].value = value
        self.pvDB[reason].time = cas.epicsTimeStamp()
        if self.pvDB[reason].mask:
            self.pvDB[reason].flag = True
        # check whether alarm/severity update is needed
        alarm, severity = pv.info.checkAlarm(value)
        self.setParamStatus(reason, alarm, severity)
        logging.getLogger('pcaspy.Driver.setParam')\
            .debug('%s: %s', reason, self.pvDB[reason])

    def setParamStatus(self, reason, alarm=None, severity=None):
        """set PV status and severity and request update

        :param str reason: PV base name
        :param alarm: alarm state
        :param severity: severity state

        The PVs' alarm status and severity are automatically set in :meth:`setParam`.
        If the status and severity need to be set explicitly to override the defaults, :meth:`setParamStatus` must
        be called *after* :meth:`setParam`.

        The new alarm status/severity will be pushed to registered clients the next time when :meth:`updatePVs` is called.
        """
        if alarm is not None and self.pvDB[reason].alarm != alarm:
            self.pvDB[reason].alarm = alarm
            self.pvDB[reason].mask |= cas.DBE_ALARM
            self.pvDB[reason].flag = True
        if severity is not None and self.pvDB[reason].severity != severity:
            self.pvDB[reason].severity = severity
            self.pvDB[reason].mask |= cas.DBE_ALARM
            self.pvDB[reason].flag = True

    def setParamEnums(self, reason, enums, states=None):
        """ set PV enumerate strings and severity states

        :param str reason: PV base name
        :param list enums: string representation of the enumerate states
        :param list states: alarm severity of the enumerate states.

        The number of elements in *states* must match that of *enums*.
        If *None* is given, the list is populated with *Severity.NO_ALARM*.

        The new enumerate strings will be pushed to registered clients the next time when :meth:`updatePVs` is called.

        .. note:: The monitoring client needs to use *DBR_GR_XXX* or *DBR_CTRL_XXX* request type and *DBE_PROPERTY*
                  event mask when issuing the subscription. This requires EPICS base 3.14.12.6+.
        """
        if states is None:
            states = [Alarm.NO_ALARM] * len(enums)
        if len(enums) != len(states):
            raise ValueError('enums and states must have the same length')
        pv = manager.pvs[self.port][reason]
        if pv.info.enums != enums:
            pv.info.enums = enums
            pv.info.states = states
            self.pvDB[reason].mask |= cas.DBE_PROPERTY
            self.pvDB[reason].flag = True

    def setParamInfo(self, reason, info):
        """
        set PV meta info, limits, precision, limits, units.

        :param str reason: PV base name
        :param dict info: information dictionary, same as used in :meth:`SimpleServer.createPV`.

        The new meta information will be pushed to registered clients the next time when :meth:`updatePVs` is called.

        .. note:: The monitoring client needs to use *DBR_GR_XXX* or *DBR_CTRL_XXX* request type and *DBE_PROPERTY*
                  event mask when issuing the subscription. This requires EPICS base 3.14.12.6+.
        """
        # copy new information
        pv = manager.pvs[self.port][reason]
        for k, v in info.items():
            if hasattr(pv.info, k):
                setattr(pv.info, k, v)
        pv.info.validateLimit()

        # recheck alarm
        alarm, severity = pv.info.checkAlarm(self.pvDB[reason].value)
        self.setParamStatus(reason, alarm, severity)

        # mark event mask and flag
        self.pvDB[reason].mask |= cas.DBE_PROPERTY
        self.pvDB[reason].flag = True

    def getParam(self, reason):
        """retrieve PV value

        :param str reason: PV base name
        :return: PV current value

        """
        return self.pvDB[reason].value

    def getParamDB(self, reason):
        """
        Return the PV data information

        :param str reason: PV base name
        :return: PV current data information
        :rtype: :class:`Data`

        """

        return self.pvDB[reason]

    def getParamInfo(self, reason, info_keys=None):
        """
        Get PV info fields. This function returns a dictionary with info/value pairs,
        where each entry of the info_keys-parameter results in a dictionary entry if
        the PVInfo-object has such an attribute. Attributes that do not exist are ignored.
        Valid attributes are the same as used in :meth:`SimpleServer.createPV`.

        If no info_keys are specified, all PV info keys are returned.

        :param str reason: PV base name
        :param list info_keys: List of keys for what information to obtain
        :return: Dictionary with PV info fields and their current values
        :rtype: dict
        """
        pv = manager.pvs[self.port][reason]

        if info_keys is None:
            info_keys = ['states', 'prec', 'unit', 'lolim', 'hilim',
                         'hihi', 'lolo', 'high', 'low', 'scan', 'asyn', 'adel', 'mdel',
                         'asg', 'port', 'enums', 'count', 'type', 'value']

        info_dict = {}
        for key in info_keys:
            if hasattr(pv.info, key):
                info_dict[key] = getattr(pv.info, key)

        return info_dict

    def callbackPV(self, reason):
        """Inform asynchronous write completion

        :param str reason: PV base name

        """
        pv = manager.pvs[self.port][reason]
        if pv.info.asyn:
            pv.endAsyncWrite(cas.S_casApp_success)

    def updatePVs(self):
        """Post update events on all PVs with value, alarm status or metadata changed"""
        for reason in self.pvDB:
            self.updatePV(reason)

    def updatePV(self, reason):
        """Post update event on the PV if value, alarm status or metadata changes

        :param str reason: PV base name
        """
        pv = manager.pvs[self.port][reason]
        if self.pvDB[reason].flag and pv.info.scan == 0:
            self.pvDB[reason].flag = False
            pv.updateValue(self.pvDB[reason])
            self.pvDB[reason].mask = 0


# map aitType to string representation
_ait_d = {'enum':   cas.aitEnumEnum16,
          'str':    cas.aitEnumString,
          'string': cas.aitEnumString,
          'float':  cas.aitEnumFloat64,
          'int':    cas.aitEnumInt32,
          'short':  cas.aitEnumInt16,
          'char':   cas.aitEnumUint8,
          }

# map aitType to gddAppType_dbr_ctrl_xxx
_dbr_d = {
    cas.aitEnumUint8:   32,
    cas.aitEnumInt16:   29,
    cas.aitEnumInt32:   33,
    cas.aitEnumFloat64: 34,
}


class PVInfo(object):
    def __init__(self, info):
        # initialize from info dict with defaults
        self.count = info.get('count', 1)
        self.type = _ait_d[info.get('type', 'float')]
        # check the number of enum states and
        # the state string do not exceed the maximum
        enums = info.get('enums', [])
        if len(enums) > cas.MAX_ENUM_STATES:
            sys.stderr.write('enums exceeds the maximum allowed states %d\n' % cas.MAX_ENUM_STATES)
            enums = enums[:cas.MAX_ENUM_STATES]
        self.enums = []
        for enum in enums:
            if len(enum) >= cas.MAX_ENUM_STRING_SIZE:
                sys.stderr.write('enums state "%s" exceeds the maximum length %d\n'
                                 % (enum, cas.MAX_ENUM_STRING_SIZE-1))
                enum = enum[:cas.MAX_ENUM_STRING_SIZE-1]
            self.enums.append(enum)
        self.states = info.get('states', [])
        # initialize enum severity states if not specified
        if not self.states:
            self.states = len(self.enums) * [Severity.NO_ALARM]
        self.prec = info.get('prec', 0.0)
        self.unit = info.get('unit', '')
        self.lolim = info.get('lolim', 0.0)
        self.hilim = info.get('hilim', 0.0)
        self.hihi = info.get('hihi', 0.0)
        self.lolo = info.get('lolo', 0.0)
        self.high = info.get('high', 0.0)
        self.low = info.get('low',  0.0)
        self.adel = info.get('adel', 0.0)
        self.mdel = info.get('mdel', 0.0)
        self.scan = info.get('scan', 0)
        self.asyn = info.get('asyn', False)
        self.asg = info.get('asg', '')
        self.reason = ''
        self.port = info.get('port', 'default')
        # validate alarm limit
        self.valid_low_high = False
        self.valid_lolo_hihi = False
        self.validateLimit()

        # initialize value based on type and count
        if self.type in [cas.aitEnumString, cas.aitEnumFixedString, cas.aitEnumUint8]:
            value = ''
        else:
            value = 0
        if self.count > 1 and self.type is not cas.aitEnumUint8:
            value = [value] * self.count
        self.value = info.get('value', value)
        # initialize last monitor/archive value
        self.mlst = self.value
        self.alst = self.value

    def validateLimit(self):
        # validate alarm limit
        if self.lolo >= self.hihi:
            self.valid_lolo_hihi = False
        else:
            self.valid_lolo_hihi = True

        if self.low >= self.high:
            self.valid_low_high = False
        else:
            self.valid_low_high = True

    def checkValue(self, newValue):
        """Check value change event"""
        mask = 0
        # array type always gets notified
        if self.count > 1:
            mask = (cas.DBE_LOG | cas.DBE_VALUE)
        # string type's equality is checked
        elif self.type in [cas.aitEnumString, cas.aitEnumFixedString]:
            if self.mlst != newValue:
                mask |= cas.DBE_VALUE
                self.mlst = newValue
            if self.alst != newValue:
                mask |= cas.DBE_LOG
                self.alst = newValue
        # scalar numeric type is checked against archive and monitor deadband
        else:
            if abs(self.mlst - newValue) > self.mdel:
                mask |= cas.DBE_VALUE
                self.mlst = newValue
            if abs(self.alst - newValue) > self.adel:
                mask |= cas.DBE_LOG
                self.alst = newValue
        return mask

    def checkAlarm(self, value):
        if self.type == cas.aitEnumEnum16:
            return self._checkEnumAlarm(value)
        elif self.type in [cas.aitEnumFloat64, cas.aitEnumInt32, cas.aitEnumInt16]:
            return self._checkNumericAlarm(value)
        elif self.type in [cas.aitEnumString, cas.aitEnumFixedString, cas.aitEnumUint8]:
            return Alarm.NO_ALARM, Severity.NO_ALARM
        else:
            return None, None

    def _checkNumericAlarm(self, value):
        severity = Severity.NO_ALARM
        alarm = Alarm.NO_ALARM

        if self.valid_low_high:
            if self._compareNumeric(value, self.low, operator.le):
                alarm = Alarm.LOW_ALARM
                severity = Severity.MINOR_ALARM
            elif self._compareNumeric(value, self.high, operator.ge):
                alarm = Alarm.HIGH_ALARM
                severity = Severity.MINOR_ALARM

        if self.valid_lolo_hihi:
            if self._compareNumeric(value, self.lolo, operator.le):
                alarm = Alarm.LOLO_ALARM
                severity = Severity.MAJOR_ALARM
            elif self._compareNumeric(value, self.hihi, operator.ge):
                alarm = Alarm.HIHI_ALARM
                severity = Severity.MAJOR_ALARM

        return alarm, severity

    def _checkEnumAlarm(self, value):
        if 0 <= value < len(self.states):
            severity = self.states[value]
            if severity == Severity.NO_ALARM:
                alarm = Alarm.NO_ALARM
            else:
                alarm = Alarm.STATE_ALARM
        else:
            severity = Severity.MAJOR_ALARM
            alarm = Alarm.STATE_ALARM

        return alarm, severity

    def _compareNumeric(self, value, limit, op):
        """
        Compare value and limit with comparison operator.

        :param value: numeric scalar or sequence
        :param limit: numeric scalar
        :param op: comparision operators, le, ge etc
        """
        if isinstance(value, collections.Iterable):
            return any(op(v, limit) for v in value)
        else:
            return op(value, limit)


class SimplePV(cas.casPV):
    """
    This class represent the PV entity and its associated attributes.

    It is to be created by server application on startup.
    It derives from :cpp:class:`PV` and implements the virtual methods.

    .. note:: This is considered an internal class and should not be referenced by module users.
    """
    def __init__(self, name, info):
        cas.casPV.__init__(self)
        self.name = name
        self.info = info
        self.interest = False
        if info.asg:
            self.setAccessSecurityGroup(info.asg)
        # scan thread
        if self.info.scan > 0:
            self.tid = threading.Thread(target=self.scan)
            self.tid.setDaemon(True)
            self.tid.start()

    def scan(self):
        while True:
            driver = manager.driver.get(self.info.port)
            if driver:
                # read value from driver and write to driver's param database
                newValue = driver.read(self.info.reason)
                driver.setParam(self.info.reason, newValue)
                # post update events if necessary
                dbValue = driver.getParamDB(self.info.reason)
                if dbValue.flag:
                    dbValue.flag = False
                    self.updateValue(dbValue)
                    dbValue.mask = 0
            time.sleep(self.info.scan)

    def interestRegister(self):
        self.interest = True
        return cas.S_casApp_success

    def interestDelete(self):
        self.interest = False

    def writeValue(self, gddValue):
        # get driver object
        driver = manager.driver.get(self.info.port)
        if not driver:
            logging.getLogger('pcaspy.SimplePV.writeValue').\
                warning('%s: No driver is registered for port %s', self.info.reason, self.info.port)
            return cas.S_casApp_undefined
        # call out driver support
        success = driver.write(self.info.reason, gddValue.get())
        if success is False:
            logging.getLogger('pcaspy.SimplePV.writeValue').\
                warning('%s: Driver rejects value %s', self.info.reason, gddValue.get())
            driver.setParamStatus(self.info.reason, Alarm.WRITE_ALARM, Severity.INVALID_ALARM)
        driver.updatePV(self.info.reason)
        return success

    def write(self, context, value):
        # delegate asynchronous write to python writeNotify method
        # only if writeNotify not present in C++ library
        if not cas.EPICS_HAS_WRITENOTIFY and self.info.asyn:
            return self.writeNotify(context, value)
        else:
            self.writeValue(value)
            return cas.S_casApp_success

    def writeNotify(self, context, value):
        # postpone request if one already in process
        if self.hasAsyncWrite():
            return cas.S_casApp_postponeAsyncIO

        # do asynchronous only if PV supports
        if self.info.asyn:
            # register async write io
            self.startAsyncWrite(context)
            # call out driver
            success = self.writeValue(value)
            # if not successful, clean the async write io
            # pass status S_cas_success instead of cas.S_casApp_canceledAsyncIO
            # so that client wont see error message.
            if not success:
                self.endAsyncWrite(cas.S_cas_success)
            # server library expects status S_casApp_asynCompletion if async write io has been initiated.
            return cas.S_casApp_asyncCompletion
        else:
            # call out driver
            success = self.writeValue(value)
            return cas.S_casApp_success

    def updateValue(self, dbValue):
        if not self.interest:
            return

        gddValue = cas.gdd(16, self.info.type)  # gddAppType_value
        if self.info.count > 1:
            gddValue.setDimension(1)
            gddValue.setBound(0, 0, self.info.count)
        gddValue.put(dbValue.value)
        gddValue.setTimeStamp(dbValue.time)
        gddValue.setStatSevr(dbValue.alarm, dbValue.severity)

        if self.info.type == cas.aitEnumEnum16:
            gddCtrl = cas.gdd.createDD(31)  # gddAppType_dbr_ctrl_enum
            gddCtrl[1].put(gddValue)
            gddCtrl[2].put(self.info.enums)
        elif self.info.type == cas.aitEnumString:  # string type has no control info
            gddCtrl = gddValue
        else:
            gddCtrl = cas.gdd.createDD(_dbr_d[self.info.type])  # gddAppType_dbr_ctrl_xxx
            gddCtrl[1].put(self.info.unit)
            gddCtrl[2].put(self.info.low)
            gddCtrl[3].put(self.info.high)
            gddCtrl[4].put(self.info.lolo)
            gddCtrl[5].put(self.info.hihi)
            gddCtrl[6].put(self.info.lolim)
            gddCtrl[7].put(self.info.hilim)
            gddCtrl[8].put(self.info.lolim)
            gddCtrl[9].put(self.info.hilim)
            if self.info.type == cas.aitEnumFloat64:
                gddCtrl[10].put(self.info.prec)
                gddCtrl[11].put(gddValue)
            else:
                gddCtrl[10].put(gddValue)

        self.postEvent(dbValue.mask, gddCtrl)

    def getValue(self, value):
        # get driver object
        driver = manager.driver.get(self.info.port)
        if not driver:
            logging.getLogger('pcaspy.SimplePV.getValue')\
                .warning('%s: No driver is registered for port %s', self.info.reason, self.info.port)
            return cas.S_casApp_undefined
        # set gdd type if necessary
        if value.primitiveType() == cas.aitEnumInvalid:
            value.setPrimType(self.info.type)
        # set gdd value
        if self.info.scan > 0:
            newValue = driver.getParam(self.info.reason)
        else:
            newValue = driver.read(self.info.reason)
        if newValue is None:
            logging.getLogger('pcaspy.SimplePV.getValue')\
                .warning('%s: Driver returns None', self.info.reason)
            return cas.S_casApp_undefined
        logging.getLogger('pcaspy.SimplePV.getValue')\
            .debug('%s: Read value %s', self.info.reason, newValue)
        value.put(newValue)
        # set gdd info
        dbValue = driver.getParamDB(self.info.reason)
        value.setStatSevr(dbValue.alarm, dbValue.severity)
        value.setTimeStamp(dbValue.time)
        return cas.S_casApp_success

    def getPrecision(self, prec):
        prec.put(self.info.prec)
        return cas.S_casApp_success

    def getUnits(self, unit):
        unit.put(self.info.unit)
        return cas.S_casApp_success

    def getEnums(self, enums):
        if self.info.enums:
            enums.put(self.info.enums)
        return cas.S_casApp_success

    def getHighLimit(self, hilim):
        hilim.put(self.info.hilim)
        return cas.S_casApp_success

    def getLowLimit(self, lolim):
        lolim.put(self.info.lolim)
        return cas.S_casApp_success

    def getHighAlarmLimit(self, hilim):
        hilim.put(self.info.hihi)
        return cas.S_casApp_success

    def getLowAlarmLimit(self, lolim):
        lolim.put(self.info.lolo)
        return cas.S_casApp_success

    def getHighWarnLimit(self, hilim):
        hilim.put(self.info.high)
        return cas.S_casApp_success

    def getLowWarnLimit(self, lolim):
        lolim.put(self.info.low)
        return cas.S_casApp_success

    def bestExternalType(self):
        return self.info.type

    def maxDimension(self):
        if self.info.count > 1:
            return 1
        else:
            return 0

    def maxBound(self, dims):
        if dims == 0:
            return self.info.count
        else:
            return 0

    def getName(self):
        return self.name


class SimpleServer(cas.caServer):
    """
    This class encapsulates transactions performed by channel access server.
    It stands between the channel access client and the driver object.
    It answers the basic channel access discover requests and forwards the
    read/write requests to driver object.

    It derives from :cpp:class:`caServer`. In addition to implement the virtual methods,
    it adds method :meth:`createPV` to create the PVs and :meth:`process` to process server requests.
    ::

        server = SimpleServer()
        server.createPV(prefix, pvdb)
        while True:
            server.process(0.1)

    """
    def __init__(self):
        cas.caServer.__init__(self)

    def __del__(self):
        cas.asCaStop()

    def pvExistTest(self, context, addr, fullname):
        if fullname in manager.pvf:
            return cas.pverExistsHere
        else:
            return cas.pverDoesNotExistHere

    def pvAttach(self, context, fullname):
        return manager.pvf.get(fullname, cas.S_casApp_pvNotFound)

    @staticmethod
    def createPV(prefix, pvdb):
        """
        Create PV based on prefix and database definition pvdb

        :param str prefix:          Name prefixing the *base_name* defined in *pvdb*
        :param dict pvdb:           PV database configuration

        pvdb is a Python *dict* assuming the following format,
        ::

          pvdb = {
            'base_name' : {
              'field_name' : value,
            },
          }

        The base_name is unique and will be prefixed to create PV full name.
        This PV configuration is expressed again in a dict. The *field_name*
        is used to configure the PV properties.

        .. _database-field-definition:
        .. table:: Database Field Definition

          ========  =======    ===============================================
          Field     Default    Description
          ========  =======    ===============================================
          type      'float'    PV data type. enum, string, char, float or int
          count     1          Number of elements
          enums     []         String representations of the enumerate states
          states    []         Severity values of the enumerate states.
                               Any of the following, Severity.NO_ALARM, Severity.MINOR_ALARM,
                               Severity.MAJOR_ALARM, Severity.INVALID_ALARM
          prec      0          Data precision
          unit      ''         Physical meaning of data
          lolim     0          Data low limit for graphics display
          hilim     0          Data high limit for graphics display
          low       0          Data low limit for alarm
          high      0          Data high limit for alarm
          lolo      0          Data low low limit for alarm
          hihi      0          Data high high limit for alarm
          adel      0          Archive deadband
          mdel      0          Monitor, value change deadband
          scan      0          Scan period in second. 0 means passive
          asyn      False      Process finishes asynchronously if True
          asg       ''         Access security group name
          value     0 or ''    Data initial value
          ========  =======    ===============================================

        The data type supported has been greatly reduced from C++ PCAS to match Python native types.
        Numeric types are 'float' and 'int', corresponding to DBF_DOUBLE and DBF_LONG of EPICS IOC.
        The display limits are defined by *lolim* abd *hilim*.
        The alarm limits are defined by *low*, *high*, *lolo*, *hihi*.

        Fixed width string, 40 characters as of EPICS 3.14, is of type 'string'.
        Long string is supported using 'char' type and specify the *count* field large enough.
        'enum' type defines a list of choices by *enums* field, and optional associated severity by *states*.

        The *adel* and *mdel* fields specify the deadbands to trigger an archive and value change event respectively.

        *asyn* if set to be True. Any channel access client write with callback option, i.e. calling `ca_put_callback`,
        will be noticed only when :meth:`Driver.callbackPV` being called.

        """
        for basename, pvinfo in pvdb.items():
            pvinfo = PVInfo(pvinfo)
            pvinfo.reason = basename
            pvinfo.name = prefix + basename
            pv = SimplePV(pvinfo.name, pvinfo)
            manager.pvf[pvinfo.name] = pv
            if pvinfo.port not in manager.pvs:
                manager.pvs[pvinfo.port] = {}
            manager.pvs[pvinfo.port][basename] = pv

    @staticmethod
    def initAccessSecurityFile(filename, **subst):
        """
        Load access security configuration file

        :param str filename:    Name of the access security configuration file
        :param subst:           Substitute macros specified by keyword arguments

        .. note::
            This must be called before :meth:`createPV`.

        """
        macro = ','.join(['%s=%s' % (k, v) for k, v in subst.items()])
        cas.asInitFile(filename, macro)
        cas.asCaStart()

    @staticmethod
    def process(delay):
        """
        Process server transactions.

        :param float delay: Processing time in second

        This method should be called so frequent so that the incoming channel access
        requests are answered in time. Normally called in the loop::

            server = SimpleServer()
            ...
            while True:
                server.process(0.1)


        """
        cas.process(delay)



#!/usr/bin/env python

from pcaspy import Driver, SimpleServer, Severity

prefix = 'MTEST:'
pvdb = {
    'ENUM' : {
        'type' : 'enum',
        'enums': ['ZERO', 'ONE']
    },
    'RAND' : {
    },
    'RAND.PREC': {
        'type' : 'int'
    },
    'RAND.EGU' : {
        'type' : 'str'
    },
    'CHANGE' : {
        'type' : 'int',
        'value': 2,
    }
}

enums = ['ZERO', 'ONE', 'TWO', 'THREE', 'FOUR', 'FIVE',
         'SIX', 'SEVEN', 'EIGHT', 'NINE', 'TEN']

class myDriver(Driver):
    def  __init__(self):
        super(myDriver, self).__init__()

    def write(self, reason, value):
        status = True
        if reason == 'CHANGE':
            num_states = value
            if num_states >= 1 and num_states <= 11:
                status = True
                self.setParamEnums('ENUM', enums[:num_states])
            else:
                status = False
        elif reason == 'RAND.EGU':
            self.setParamInfo('RAND', {'unit': value})
        elif reason == 'RAND.PREC':
            self.setParamInfo('RAND', {'prec': value})

        if status:
            self.setParam(reason, value)
        return status

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.DEBUG)

    server = SimpleServer()
    server.createPV(prefix, pvdb)
    driver = myDriver()

    # process CA transactions
    while True:
        server.process(0.1)



#!/usr/bin/env python
import time
import sys
import threading
import subprocess
import shlex

from pcaspy import Driver, SimpleServer

prefix = 'MTEST:'
pvdb = { 
    'COMMAND' : {
        'type' : 'char',
        'count': 128,
        'asyn' : True
    },
    'OUTPUT'  : {
        'type' : 'char',
        'count': 500,
    },
    'STATUS'  : {
        'type' : 'enum',
        'enums': ['DONE', 'BUSY']
    },
    'ERROR'   : {
        'type' : 'string',
    },
}
import math
class myDriver(Driver):
    def __init__(self):
        Driver.__init__(self)
        self.tid = None 

    def write(self, reason, value):
        status = True
        # take proper actions
        if reason == 'COMMAND':
            if not self.tid:
                command = value
                if command:
                    self.tid = threading.Thread(target=self.runShell,args=(command,))
                    self.tid.start()
            else:
                status = False
        else:
            status = False
        # store the values
        if status:
            self.setParam(reason, value)
        return status

    def runShell(self, command):
        print("DEBUG: Run ", command)
        # set status BUSY
        self.setParam('STATUS', 1)
        self.updatePVs()
        # run shell
        try:
            time.sleep(0.01)
            proc = subprocess.Popen(shlex.split(command), 
                    stdout = subprocess.PIPE, 
                    stderr = subprocess.PIPE)
            proc.wait()
        except OSError:
            self.setParam('ERROR', str(sys.exc_info()[1]))
            self.setParam('OUTPUT', '')
        else:
            self.setParam('ERROR', proc.stderr.read().rstrip())
            self.setParam('OUTPUT', proc.stdout.read().rstrip())
        self.callbackPV('COMMAND')
        # set status DONE
        self.setParam('STATUS', 0)
        self.updatePVs()
        self.tid = None
        print("DEBUG: Finish ", command)

if __name__ == '__main__':
    server = SimpleServer()
    server.createPV(prefix, pvdb)
    driver = myDriver()

    while True:
        # process CA transactions
        server.process(0.1)



from pcaspy.cas import gdd
import time

while True:
    gddValue = gdd()
    gddValue.put(range(1000000))
    v = gddValue.get()
    gddCtrl = gdd.createDD(34) # gddAppType_dbr_ctrl_double
    gddCtrl[1].put('eV')
    gddCtrl[2].put(0)
    gddCtrl[3].put(1)
    gddCtrl[4].put(0)
    gddCtrl[5].put(1)
    gddCtrl[6].put(0)
    gddCtrl[7].put(1)
    gddCtrl[8].put(0)
    gddCtrl[9].put(1)
    gddCtrl[10].put(3)
    gddCtrl[11].put(gddValue)


    time.sleep(0.1)



#!/usr/bin/env python

from pcaspy import SimpleServer, Driver
import random

prefix = 'MTEST:'
pvdb = {
    'RAND' : {
        'prec' : 3,
        'scan' : 1,
        'count': 10,
    },
}

class myDriver(Driver):
    def __init__(self):
        super(myDriver, self).__init__()

    def read(self, reason):
        if reason == 'RAND':
            value = [random.random() for i in range(10)]
        else:
            value = self.getParam(reason)
        return value

if __name__ == '__main__':
    server = SimpleServer()
    server.createPV(prefix, pvdb)
    driver = myDriver()

    # process CA transactions
    while True:
        server.process(0.1)



from pcaspy import cas

try:
    import numpy
except ImportError:
    print('Skipping numpy tests')
    numpy = None

import unittest

class TestGDD(unittest.TestCase):
    def setUp(self):
        self.s = cas.gdd()

    def test_copy(self):
        d = cas.gdd()
        d.put([1,2,3,4])
        self.s.put(d)
        self.assertEqual(self.s.get(), [1,2,3,4])

    def test_string(self):
        self.s.put("sdcsd")
        self.assertEqual(self.s.get(), "sdcsd")

    def test_numeric(self):
        self.s.put(12)
        self.assertEqual(self.s.get(), 12)

    def test_string_array(self):
        self.s.put(["sdcsd","sdcsd"])
        self.assertEqual(self.s.get(), ["sdcsd","sdcsd"])

    def test_numeric_array(self):
        self.s.put(range(2))
        self.assertEqual(self.s.get(), [0,1])

    def test_numpy_scaler(self):
        if numpy is None:
            return
        self.s.put(numpy.int32(12))
        self.assertEqual(self.s.get(), 12)

    def test_numpy_array(self):
        if numpy is None:
            return
        self.s.put(numpy.arange(2, dtype='i4'))
        self.assertEqual(self.s.get(), [0, 1])

    def test_numpy_unsupported_dtype(self):
        if numpy is None:
            return
        self.s.put(numpy.arange(2, dtype='i8'))
        self.assertEqual(self.s.get(), [0, 1])

    def test_dim_enlarge(self):
        self.s.put(1)
        self.assertEqual(self.s.get(), 1)

        self.s.put(range(2))
        self.assertEqual(self.s.get(), [0, 1])

    def test_convert_to_numeric(self):
        self.s.put(1)
        self.assertEqual(self.s.get(), 1)

        self.s.put('2')
        self.assertEqual(self.s.get(), 2)

        self.s.put('qwf')
        self.assertEqual(self.s.get(), 2)

    def test_convert_to_string(self):
        self.s.put('sdcsd')
        self.assertEqual(self.s.get(), 'sdcsd')

        self.s.put(2)
        self.assertEqual(float(self.s.get()), 2)

    def test_string_to_char_array(self):
        self.s.setPrimType(cas.aitEnumUint8)
        self.s.setDimension(1)
        self.s.setBound(0, 0, 10)
        self.s.put('sdcsd')
        self.assertEqual(self.s.get(), 'sdcsd')
        self.s.put('')
        self.assertEqual(self.s.get(), '')

    def test_string_to_numeric_scalar(self):
        self.s.setPrimType(cas.aitEnumUint16)
        self.s.put('sc') # no conversion, 0
        self.assertEqual(self.s.get(), 0)
        self.s.put('a')  # heximal 10
        self.assertEqual(self.s.get(), 10)
        self.s.put('10') # decimal 10
        self.assertEqual(self.s.get(), 10)

    def tearDown(self):
        del self.s

if __name__ == '__main__':
    unittest.main()



from pcaspy import Driver, SimpleServer
import pcaspy.cas as cas

import re

prefix = 'MTEST:'
pvdb = { 'STATIC' : {} }

# keep reference to dynamic spectra pvs
spvs = {}

class SpectrumSimulator(object):
    """
    An object to produce spectrum on request.
    """
    def get(self, index):
        """
        :param int index: the spectrum id
        :return: a list of 100 integers
        """
        return [index] * 100

class SpectrumPV(cas.casPV):
    def __init__(self, name):
        cas.casPV.__init__(self)
        self.name = name
        self.index = int(re.match('MTEST:SPECTRUM(\d+)', name).group(1))

    def getValue(self, value):
        value.put([self.index]*100)
        return cas.S_casApp_success

    def maxDimension(self):
        return 1

    def maxBound(self, dims):
        return 100


# The driver serves normal static PVs
class myDriver(Driver):
    def __init__(self):
        Driver.__init__(self)


class myServer(SimpleServer):

    def __init__(self):
        SimpleServer.__init__(self)

    def pvExistTest(self, context, addr, fullname):
        if fullname.startswith('MTEST:SPECTRUM'):
            return cas.pverExistsHere
        else:
            return SimpleServer.pvExistTest(self, context, addr, fullname)

    def pvAttach(self, context, fullname):
        if fullname.startswith('MTEST:SPECTRUM'):
            if fullname not in spvs:
                pv = SpectrumPV(fullname)
                spvs[fullname] = pv
            return spvs[fullname]
        else:
            return SimpleServer.pvAttach(self, context, fullname)


if __name__ == '__main__':
    server = myServer()
    server.createPV(prefix, pvdb)
    driver = myDriver()

    while True:
        server.process(1)



#!/usr/bin/env python

from pcaspy import Driver, SimpleServer
import time

prefix = 'MTEST:'
pvdb = {
    'RAND' : {
        'prec' : 3,
        'count': 3,
    },
}

import threading
import numpy

class myDriver(Driver):
    def  __init__(self):
        super(myDriver, self).__init__()
        self.value = numpy.array([1,2,3])
        tid = threading.Thread(target = self.do)
        tid.setDaemon(True)
        tid.start()

    def read(self, reason):
        foo
        pass

    def write(self, reason, value):
        pass

    def do(self,):
        while True:
            self.value[1] += 1
            self.setParam('RAND', self.value)
            self.updatePVs()
            time.sleep(1)

if __name__ == '__main__':
    server = SimpleServer()
    server.createPV(prefix, pvdb)
    driver = myDriver()

    # process CA transactions
    while True:
        server.process(0.1)



#!/usr/bin/env python

from pcaspy import Driver, SimpleServer, Alarm, Severity

prefix = 'MTEST:'
pvdb = {
    'RAND' : {
        'prec' : 3,
        'low' : -5, 'high': 5,
        'lolo': -10,'hihi': 10,
    },
    'ARRAY': {
        'count': 30,
        'value': 30*[1],
        'low': -5, 'high': 5,
        'lolo': -10, 'hihi': 10
    },
    'STATUS' : {
        'type' : 'enum',
        'enums':  ['OK', 'ERROR'],
        'states': [Severity.NO_ALARM, Severity.MAJOR_ALARM]
    },
    'MSG' : {
        'type' : 'str',
    }
}

class myDriver(Driver):
    def  __init__(self):
        super(myDriver, self).__init__()

    def write(self, reason, value):
        status = False
        if reason == 'MSG':
            status = True
            # store the value and this also resets alarm status and severity for string type
            self.setParam(reason, value)
            # set alarm status and severity
            if value != '':
                self.setParamStatus(reason, Alarm.COMM_ALARM, Severity.MINOR_ALARM)
            else:
                self.setParamStatus(reason, Alarm.NO_ALARM, Severity.NO_ALARM)
        else:
            status = True
            # store the value and this also resets alarm status and severity for string type
            self.setParam(reason, value)

        return status

if __name__ == '__main__':
    server = SimpleServer()
    server.createPV(prefix, pvdb)
    driver = myDriver()

    # process CA transactions
    while True:
        server.process(0.1)



import threading


# Thread running server processing loop
class ServerThread(threading.Thread):
    """
    A helper class to run server in a thread.

    The following snippet runs the server for 4 seconds and quit::

        server = SimpleServer()
        server_thread = ServerThread(server)
        server_thread.start()
        time.sleep(4)
        server_thread.stop()

    """
    def __init__(self, server):
        """
        :param server: :class:`pcaspy.SimpleServer` object
        """
        super(ServerThread, self).__init__()
        self.server = server
        self.running = True

    def run(self):
        """
        Start the server processing
        """
        while self.running:
            self.server.process(0.1)

    def stop(self):
        """
        Stop the server processing
        """
        self.running = False



#!/usr/bin/env python
"""
This trivial program snoops all PV search requests and prints them out.
Its only purpose is to demonstrate how channel access disconvers PVs.
"""
import pcaspy.cas as cas
import logging
logging.basicConfig(level=logging.DEBUG, format='%(asctime)s: %(message)s')

class SnooperServer(cas.caServer):
    def __init__(self):
        cas.caServer.__init__(self)

    def __del__(self):
        cas.asCaStop()

    def pvExistTest(self, context, addr, fullname):
        logging.debug('%s\t%s', addr.stringConvert(1024), fullname)
        return cas.pverDoesNotExistHere

    def pvAttach(self, context, fullname):
        return cas.S_casApp_pvNotFound

    def process(self, second):
        cas.process(second)

if __name__ == '__main__':
    snooper = SnooperServer()
    while True:
        snooper.process(0.1)



#!/usr/bin/env python

"""
setup.py file for pcaspy
"""
import os
import platform
import sys
import shutil
import subprocess
import filecmp

# Use setuptools to include build_sphinx, upload/sphinx commands
try:
    from setuptools import setup, Extension
except:
    from distutils.core import setup, Extension

from distutils.command.build_py import build_py as _build_py

# build_py runs before build_ext so that swig generated module is not copied
# See http://bugs.python.org/issue7562
# This is a workaround to run build_ext ahead of build_py
class build_py(_build_py):
    def run(self):
        self.run_command('build_ext')
        _build_py.run(self)

# python 2/3 compatible way to load module from file
def load_module(name, location):
    if sys.hexversion < 0x03050000:
        import imp
        module = imp.load_source(name, location)
    else:
        import importlib
        spec = importlib.util.spec_from_file_location(name, location)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    return module

# define EPICS base path and host arch
EPICSBASE = os.environ.get("EPICS_BASE")
HOSTARCH = os.environ.get("EPICS_HOST_ARCH")
SHARED = os.environ.get("EPICS_SHARED")
# guess from EPICS root environment variable
if not EPICSBASE:
    EPICSROOT = os.environ.get("EPICS")
    if EPICSROOT:
        EPICSBASE = os.path.join(EPICSROOT, 'base')

if not EPICSBASE or not os.path.exists(EPICSBASE) or not HOSTARCH:
    raise IOError("Please define/validate EPICS_BASE environment variable")

# check EPICS version
PRE315 = True
if os.path.exists(os.path.join(EPICSBASE, 'include', 'compiler')):
    PRE315 = False

# common libraries to link
libraries = ['cas', 'ca', 'gdd', 'Com']
if PRE315:
    libraries.insert(0, 'asIoc')
else:
    libraries.insert(0, 'dbCore')
umacros = []
macros   = []
cflags = []
lflags = []
dlls = []
extra_objects = []
# platform dependent libraries and macros
UNAME = platform.system()
if  UNAME.find('CYGWIN') == 0:
    UNAME = "cygwin32"
    CMPL = 'gcc'
elif UNAME == 'Windows':
    UNAME = 'WIN32'
    # MSVC compiler
    static = False
    if HOSTARCH in ['win32-x86', 'windows-x64', 'win32-x86-debug', 'windows-x64-debug']:
        dlls = ['cas.dll', 'ca.dll', 'gdd.dll', 'Com.dll']
        if PRE315:
            dlls += ['dbIoc.dll', 'dbStaticIoc.dll', 'asIoc.dll']
        else:
            dlls += ['dbCore.dll']
        for dll in dlls:
            dllpath = os.path.join(EPICSBASE, 'bin', HOSTARCH, dll)
            if not os.path.exists(dllpath):
                static = True
                break
            dll_dest = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'pcaspy', dll)
            if not os.path.exists(dll_dest) or not filecmp.cmp(dllpath, dll_dest):
                shutil.copy(dllpath, dll_dest)
        macros += [('_CRT_SECURE_NO_WARNINGS', 'None'), ('_CRT_NONSTDC_NO_DEPRECATE', 'None'), ('EPICS_CALL_DLL', '')]
        cflags += ['/Z7']
        CMPL = 'msvc'
    if HOSTARCH in ['win32-x86-static', 'windows-x64-static'] or static:
        libraries += ['ws2_32', 'user32', 'advapi32']
        macros += [('_CRT_SECURE_NO_WARNINGS', 'None'), ('_CRT_NONSTDC_NO_DEPRECATE', 'None'), ('EPICS_DLL_NO', '')]
        umacros+= ['_DLL']
        cflags += ['/EHsc', '/Z7']
        lflags += ['/LTCG']
        if HOSTARCH[-5:] == 'debug':
            libraries += ['msvcrtd']
            lflags += ['/NODEFAULTLIB:libcmtd.lib']
        else:
            libraries += ['msvcrt']
            lflags += ['/NODEFAULTLIB:libcmt.lib']
        CMPL = 'msvc'
    # GCC compiler
    if HOSTARCH in ['win32-x86-mingw', 'windows-x64-mingw']:
        macros += [('_MINGW', ''), ('EPICS_DLL_NO', '')]
        lflags += ['-static',]
        CMPL = 'gcc'
    if HOSTARCH == 'windows-x64-mingw':
        macros += [('MS_WIN64', '')]
        CMPL = 'gcc'
elif UNAME == 'Darwin':
    CMPL = 'clang'
    if not SHARED:
        extra_objects = [os.path.join(EPICSBASE, 'lib', HOSTARCH, 'lib%s.a'%lib) for lib in libraries]
        libraries = []
elif UNAME == 'Linux':
    if not SHARED:
        extra_objects = [os.path.join(EPICSBASE, 'lib', HOSTARCH, 'lib%s.a'%lib) for lib in libraries]
        # necessary when EPICS is statically linked
        libraries = ['rt']
        if subprocess.call('nm -u %s | grep -q rl_' % os.path.join(EPICSBASE, 'lib', HOSTARCH, 'libCom.a'), shell=True) == 0:
            libraries += ['readline']
    CMPL = 'gcc'
elif UNAME == 'SunOS':
    # OS_CLASS used by EPICS
    UNAME = 'solaris'
    CMPL = 'solStudio'
else:
    raise IOError("Unsupported OS {0}".format(UNAME))

cas_module = Extension('pcaspy._cas',
                       sources  =[os.path.join('pcaspy','casdef.i'),
                                  os.path.join('pcaspy','pv.cpp'),
                                  os.path.join('pcaspy','channel.cpp'),],
                       swig_opts=['-c++','-threads','-nodefaultdtor','-I%s'% os.path.join(EPICSBASE, 'include')],
                       extra_compile_args=cflags,
                       include_dirs = [ os.path.join(EPICSBASE, 'include'),
                                        os.path.join(EPICSBASE, 'include', 'os', UNAME),
                                        os.path.join(EPICSBASE, 'include', 'compiler', CMPL)],
                       library_dirs = [ os.path.join(EPICSBASE, 'lib', HOSTARCH),],
                       libraries = libraries,
                       extra_link_args = lflags,
                       extra_objects = extra_objects,
                       define_macros = macros,
                       undef_macros  = umacros,)
# use runtime library path option if linking share libraries on *NIX
if UNAME not in ['WIN32'] and SHARED:
    cas_module.runtime_library_dirs += [os.path.join(EPICSBASE, 'lib', HOSTARCH)]

long_description = open('README.rst').read()
_version = load_module('_version', 'pcaspy/_version.py')

dist = setup (name = 'pcaspy',
              version = _version.__version__,
              description = """Portable Channel Access Server in Python""",
              long_description = long_description,
              author      = "Xiaoqiang Wang",
              author_email= "xiaoqiangwang@gmail.com",
              url         = "https://pypi.python.org/pypi/pcaspy",
              ext_modules = [cas_module],
              packages    = ["pcaspy"],
              package_data={"pcaspy" : dlls},
              cmdclass    = {'build_py':build_py},
              license     = "BSD",
              platforms   = ["Windows","Linux", "Mac OS X"],
              classifiers = ['Development Status :: 4 - Beta',
                             'Environment :: Console',
                             'Intended Audience :: Developers',
                             'License :: OSI Approved :: BSD License',
                             'Programming Language :: C++',
                             'Programming Language :: Python :: 2',
                             'Programming Language :: Python :: 3',
                             ],
              )

# Re-run the build_py to ensure that swig generated py files are also copied
build_py = build_py(dist)
build_py.ensure_finalized()
build_py.run()



#!/usr/bin/env python
"""
An example server for RaspberryPi GPIO control/monitor.

One output and one input are defined using ``PIN_OUT`` and ``PIN_INP``.
The output can either output constant low/high ('OUT'), or output a pulse of 100ms width ('TRIG').

"""

from pcaspy import Driver, SimpleServer
import RPi.GPIO as GPIO
import time

prefix = 'MTEST:'
pvdb = {
    'OUT' : {
        'type':  'enum',
        'enums': ['LOW', 'HIGH'],
    },
    'TRIG' : {
        'type':  'enum',
        'enums': ['NONE', 'TRIG'],
    },
    'PWM' : {
        'type':  'enum',
        'enums': ['STOP', 'START'],
    },
    'DC' : {
        'type':  'int',
    },
    'FREQ' : {
        'value': 1
    },
    'INP' : {
        'type':  'enum',
        'enums': ['LOW', 'HIGH'],
    },
    'TRIGED': {
        'type':  'enum',
        'enums': ['NONE', 'RISE', 'FALL'],
    }
}

PIN_OUT = 12 #: pin for output
PIN_PWM = 16 #: pin for PWM
PIN_INP = 3 #: pin for input

class myDriver(Driver):
    def  __init__(self):
        super(myDriver, self).__init__()
        GPIO.setmode(GPIO.BOARD)
        GPIO.setup(PIN_INP, GPIO.IN)
        GPIO.setup(PIN_OUT, GPIO.OUT)
        GPIO.setup(PIN_PWM, GPIO.OUT)
        GPIO.add_event_detect(PIN_INP, GPIO.BOTH, callback=self.input_callback)
        self.pwm = GPIO.PWM(PIN_PWM, 1) # 1 Hz

    def input_callback(self, channel):
        level = GPIO.input(channel)
        self.setParam('INP', level)
        if level == 0:
            self.setParam('TRIGED', 2)
        elif level == 1:
            self.setParam('TRIGED', 1)
        self.updatePVs()

    def write(self, reason, value):
        status = True

        if reason == 'OUT':
            GPIO.output(PIN_OUT, value)
        elif reason == 'TRIG':
            GPIO.output(PIN_OUT, GPIO.HIGH)
            time.sleep(0.1)
            GPIO.output(PIN_OUT, GPIO.LOW)
            value = 0
        elif reason == 'PWM':
            if value == 0:
                self.pwm.stop()
            else:
                self.pwm.start(self.getParam('DC'))
        elif reason == 'DC':
            self.pwm.ChangeDutyCycle(value)
        elif reason == 'FREQ':
            self.pwm.ChangeFrequency(value)
        else:
            status = False

        if status:
            self.setParam(reason, value)
        return status


if __name__ == '__main__':
    server = SimpleServer()
    server.createPV(prefix, pvdb)
    driver = myDriver()

    # process CA transactions
    while True:
        server.process(0.1)

